from dataclasses import dataclass
from typing import TYPE_CHECKING

from .schema import GraphObjectType

if TYPE_CHECKING:
    from ..databases.ingest_strategy import IngestionStrategy


@dataclass(frozen=True, slots=True)
class KeyIndex:
    """Defines an Index that is used as a Key on a Node Type.

    `KeyIndex`es are created by Interpretations that attempt to create and use
    a node reference as a Primary Key. Usually this implies the creation of a
    uniqueness constraint in the database.

    See information on your Underlying Graph Database Adapter for information
    on how this is implemented.
    """

    type: str
    identity_keys: frozenset

    async def ingest(self, strategy: "IngestionStrategy"):
        await strategy.upsert_key_index(self)


@dataclass(frozen=True, slots=True)
class FieldIndex:
    """Defines an index that is used on a field for a Graph Object.

    `FieldIndex`es are created on fields that do not imply a uniqueness constraint and "just" need
    to be indexes for query performance.
    """

    type: str
    field: str
    object_type: GraphObjectType

    async def ingest(self, strategy: "IngestionStrategy"):
        await strategy.upsert_field_index(self)

    @classmethod
    def for_ttl_timestamp(
        cls, type, object_type: GraphObjectType = GraphObjectType.NODE
    ):
        return cls(type, "last_ingested_at", object_type=object_type)



import json
from abc import ABC, abstractmethod
from contextlib import asynccontextmanager, contextmanager
from csv import DictReader
from glob import glob
from io import StringIO
from pathlib import Path
from tempfile import SpooledTemporaryFile
from typing import Any, AsyncGenerator, Iterable, Union

from httpx import AsyncClient

from ...model import JsonLikeDocument
from ...pluggable import Pluggable
from ...subclass_registry import SubclassRegistry
from .extractor import Extractor

SUPPORTED_FILE_FORMAT_REGISTRY = SubclassRegistry()


@SUPPORTED_FILE_FORMAT_REGISTRY.connect_baseclass
class SupportedFileFormat(Pluggable, ABC):
    def __init__(self, file: Union[Path, StringIO]) -> None:
        self.file = file

    @contextmanager
    def read_handle(self) -> StringIO:
        if isinstance(self.file, Path):
            with open(self.file, "r", encoding="utf-8") as fp:
                yield fp
        else:
            yield self.file

    def read_file(self) -> Iterable[JsonLikeDocument]:
        with self.read_handle() as fp:
            return self.read_file_from_handle(fp)

    @classmethod
    @contextmanager
    def open(cls, file: Path) -> "SupportedFileFormat":
        with open(file, "r", encoding="utf-8") as fp:
            yield cls.from_file_pointer_and_format(fp, file.suffix)

    @classmethod
    def from_file_pointer_and_format(
        cls, fp: StringIO, file_format: str
    ) -> "SupportedFileFormat":
        # Import all file formats so that they can register themselves
        cls.import_all()
        file_format = SUPPORTED_FILE_FORMAT_REGISTRY.get(file_format)
        return file_format(fp)

    @abstractmethod
    def read_file_from_handle(self, fp: StringIO) -> Iterable[JsonLikeDocument]:
        ...


class JsonFileFormat(SupportedFileFormat, alias=".json"):
    def read_file_from_handle(self, fp: StringIO) -> Iterable[JsonLikeDocument]:
        return [json.load(fp)]


class LineSeperatedJsonFileFormat(SupportedFileFormat, alias=".jsonl"):
    def read_file_from_handle(self, fp: StringIO) -> Iterable[JsonLikeDocument]:
        return (json.loads(line) for line in fp)


class TextFileFormat(SupportedFileFormat, alias=".txt"):
    def read_file_from_handle(self, fp: StringIO) -> Iterable[JsonLikeDocument]:
        return ({"line": line} for line in fp)


class CommaSeperatedValuesFileFormat(SupportedFileFormat, alias=".csv"):
    def read_file_from_handle(self, fp: StringIO) -> Iterable[JsonLikeDocument]:
        return DictReader(fp)


class FileExtractor(Extractor):
    @classmethod
    def from_file_data(cls, globs: Iterable[str]):
        all_matching_paths = (
            Path(file)
            for glob_string in globs
            for file in glob(glob_string, recursive=True)
        )
        final_paths = {p for p in all_matching_paths if p.is_file()}
        return cls(final_paths)

    def __init__(self, paths: Iterable[Path]) -> None:
        self.paths = paths

    def _ordered_paths(self) -> Iterable[Path]:
        return sorted(self.paths)

    async def extract_records(self) -> AsyncGenerator[Any, Any]:
        for path in self._ordered_paths():
            with SupportedFileFormat.open(path) as file:
                for record in file.read_file():
                    yield record


class RemoteFileExtractor(Extractor):
    def __init__(
        self, urls: Iterable[str], memory_spooling_max_size_in_mb: int = 5
    ) -> None:
        self.urls = urls
        self.memory_spooling_max_size = memory_spooling_max_size_in_mb * 1024 * 1024

    @asynccontextmanager
    async def download_file(self, client: AsyncClient, url: str) -> SupportedFileFormat:
        with SpooledTemporaryFile(max_size=self.memory_spooling_max_size) as fp:
            async with client.stream("GET", url) as response:
                async for chunk in response.aiter_bytes():
                    fp.write(chunk)
            fp.seek(0)
            yield SupportedFileFormat.from_file_pointer_and_format(fp, Path(url).suffix)

    async def extract_records(self) -> AsyncGenerator[Any, Any]:
        async with AsyncClient() as client:
            for url in self.urls:
                async with self.download_file(client, url) as file:
                    for record in file.read_file():
                        yield record



import asyncio
import traceback
from logging import getLogger
from typing import Any, AsyncGenerator, Iterable, List, Optional

from ..schema.schema import (
    AggregatedIntrospectiveIngestionComponent,
    IntrospectiveIngestionComponent,
)
from .meta import get_context
from .progress_reporter import PipelineProgressReporter
from .step import Step


async def empty_async_generator():
    for item in []:
        yield item


async def enumerate_async(iterable):
    count = 0

    async for item in iterable:
        yield count, item
        count += 1


class DoneObject:
    pass


START_EXCEPTION = "Exception in Start Process:"
WORK_BODY_EXCEPTION = "Exception in Work Body:"
STOP_EXCEPTION = "Exception in Stop Process:"


class StepException(Exception):
    """
    Exception Format:
        Exceptions in StepExecutor {exec_num} ({step_name})
            Exception in Start
                Stack
            Exception in Work Body
                Stack
            Exception in Stop
                Stack
    """

    def __init__(self, errors: dict[str, Exception], identifier: Optional[str]):
        self.exceptions = errors
        self.identifier = identifier
        super().__init__(self.build_message())

    def build_message(self):
        message = ""
        for section, exception in self.exceptions.items():
            sub_message = "".join(traceback.format_tb(exception.__traceback__))
            # Break it apart, add a tab, bring it back together.
            message += f"\t{section}\n"
            for sentence in [
                "\t\t" + sentence + "\n" for sentence in sub_message.split("\n")
            ]:
                message += sentence
        return f"Exceptions in {self.identifier}:\n" + message


class PipelineException(Exception):
    """
    Exception in Pipeline:
        Output from StepException 0
        Output from StepException 1
    """

    def __init__(self, errors: list[Exception]):
        self.errors = errors
        super().__init__(self.build_message())

    def build_message(self):
        message = ""
        sub_message = ""
        for error in self.errors:
            sub_message += error.build_message()

        # Break it apart, add a tab, bring it back together.
        for sentence in [
            "\t" + sentence + "\n" for sentence in sub_message.split("\n")
        ]:
            message += sentence

        return "Exceptions in Pipeline:\n" + message


class StepExecutor:
    def __init__(
        self,
        upstream: Optional["StepExecutor"],
        step: Step,
        outbox_size: int = 0,
        step_index: Optional[int] = 0,
        progress_reporter: Optional[PipelineProgressReporter] = None,
    ) -> None:
        self.outbox = asyncio.Queue(maxsize=outbox_size)
        self.upstream = upstream
        self.done = False
        self.step = step
        self.step_index = step_index
        self.progress_reporter = progress_reporter
        self.end_of_line = False
        self.identifier = f"{self.__class__.__name__} {self.step_index} ({self.step.__class__.__name__})"
        self.logger = getLogger(name=self.identifier)
        self.exceptions = {}

    async def outbox_generator(self):
        while not self.done or not self.outbox.empty():
            if (value := await self.outbox.get()) is not DoneObject:
                yield value
            self.outbox.task_done()

    def start(self):
        if self.progress_reporter:
            self.progress_reporter.on_start_callback()

    def set_end_of_line(self, progress_reporter: Optional[PipelineProgressReporter]):
        self.progress_reporter = progress_reporter
        self.end_of_line = True

    async def stop(self):
        self.done = True
        await self.outbox.put(DoneObject)
        await self.step.finish()

        if self.progress_reporter:
            self.progress_reporter.on_finish_callback(get_context())

    async def work_body(self):
        if self.upstream is None:
            upstream = empty_async_generator()
        else:
            upstream = self.upstream.outbox_generator()
        results = self.step.handle_async_record_stream(upstream)
        async for index, record in enumerate_async(results):
            if not self.end_of_line:
                await self.outbox.put(record)
            if self.progress_reporter:
                self.progress_reporter.report(index, record)

    def try_start(self):
        # Start only calls some callbacks on the reporter.
        # If this fails, we can just log the error and move on.
        try:
            self.start()
        except Exception as start_exception:
            self.logger.exception(
                f"Exception during start for step {self.step.__class__.__name__}. This will not be considered fatal.",
                stack_info=True,
                extra={"step": self.step.__class__.__name__},
            )
            self.exceptions[START_EXCEPTION] = start_exception

    async def try_work_body(self):
        # During the main exection of the code, we want to catch any
        # exceptions that happen and log them.
        # Since this step can no longer do any work, we will raise the
        # exception so that the pipeline can safely come to a stop.
        try:
            await self.work_body()
        except Exception as work_body_exception:
            self.logger.exception(
                f"Exception during the work body for step {self.step.__class__.__name__}. This will not be considered fatal.",
                stack_info=True,
                extra={"step": self.step.__class__.__name__},
            )
            self.exceptions[WORK_BODY_EXCEPTION] = work_body_exception

    async def try_stop(self):
        # When we're done, we need to try to call stop on the step.
        # This is because the step may have some cleanup to do.
        # If this fails, we are near the end of the exection of the pipeline
        # so we can just log the error and move on.
        try:
            await self.stop()
        except Exception as stop_exception:
            self.logger.exception(
                f"Exception during stop for step {self.step.__class__.__name__}. This will not be considered fatal.",
                stack_info=True,
                extra={"step": self.step.__class__.__name__},
            )
            self.exceptions[STOP_EXCEPTION] = stop_exception

    def raise_encountered_exceptions(self):
        if self.exceptions:
            raise StepException(errors=self.exceptions, identifier=self.identifier)

    async def work_loop(self):
        self.try_start()
        await self.try_work_body()
        await self.try_stop()
        self.raise_encountered_exceptions()


class Pipeline(AggregatedIntrospectiveIngestionComponent):
    """A pipeline is a series of steps that are executed in order."""

    __slots__ = ("steps",)

    def __init__(self, steps: List[Step], step_outbox_size: int) -> None:
        self.steps = steps
        self.step_outbox_size = step_outbox_size
        self.logger = getLogger(self.__class__.__name__)
        self.errors = []

    def build_steps_into_tasks(
        self, progress_reporter: Optional[PipelineProgressReporter] = None
    ):
        current_executor = None
        tasks = []
        for step_index, step in enumerate(self.steps):
            current_executor = StepExecutor(
                upstream=current_executor,
                step=step,
                outbox_size=self.step_outbox_size,
                step_index=step_index,
            )
            tasks.append(asyncio.create_task(current_executor.work_loop()))
        current_executor.set_end_of_line(progress_reporter)
        return tasks

    def propogate_errors_from_return_states(self, return_states):
        for return_state in return_states:
            if return_state:
                self.errors.append(return_state)

        # Raise the error and log it.
        if self.errors:
            raise PipelineException(self.errors)

    async def run(
        self, progress_reporter: Optional[PipelineProgressReporter] = None
    ) -> AsyncGenerator[Any, Any]:
        tasks = self.build_steps_into_tasks(progress_reporter)
        return_states = await asyncio.gather(*tasks, return_exceptions=True)
        self.propogate_errors_from_return_states(return_states)

    def all_subordinate_components(self) -> Iterable[IntrospectiveIngestionComponent]:
        return (s for s in self.steps if isinstance(s, IntrospectiveIngestionComponent))



from pathlib import Path

from cleo.helpers import argument

from ..operations import RunProjectCookiecutter
from .nodestream_command import NodestreamCommand
from .shared_options import DATABASE_NAME_OPTION


class New(NodestreamCommand):
    name = "new"
    description = "Generate a new nodestream project"
    arguments = [
        argument("name", "the name of the project to create"),
        argument("path", "the location of the project to create", optional=True),
    ]
    options = [DATABASE_NAME_OPTION]

    async def handle_async(self):
        project_name = self.argument("name")
        project_path = self.argument("path")
        project_path = Path(self.argument("path")) if project_path else Path.cwd()
        db = self.option("database")
        generate = RunProjectCookiecutter(project_name, project_path, db)
        await self.run_operation(generate)



from logging import getLogger
from typing import Iterable

from neo4j import AsyncDriver

from ...model import IngestionHook, Node, RelationshipWithNodes, TimeToLiveConfiguration
from ...schema.indexes import FieldIndex, KeyIndex
from ..query_executor import (
    OperationOnNodeIdentity,
    OperationOnRelationshipIdentity,
    QueryExecutor,
)
from .index_query_builder import Neo4jIndexQueryBuilder
from .ingest_query_builder import Neo4jIngestQueryBuilder
from .query import Query


class Neo4jQueryExecutor(QueryExecutor):
    def __init__(
        self,
        driver: AsyncDriver,
        ingest_query_builder: Neo4jIngestQueryBuilder,
        index_query_builder: Neo4jIndexQueryBuilder,
        database_name: str,
    ) -> None:
        self.driver = driver
        self.ingest_query_builder = ingest_query_builder
        self.index_query_builder = index_query_builder
        self.logger = getLogger(self.__class__.__name__)
        self.database_name = database_name

    async def upsert_nodes_in_bulk_with_same_operation(
        self, operation: OperationOnNodeIdentity, nodes: Iterable[Node]
    ):
        batched_query = (
            self.ingest_query_builder.generate_batch_update_node_operation_batch(
                operation, nodes
            )
        )
        await self.execute(
            batched_query.as_query(self.ingest_query_builder.apoc_iterate),
            log_result=True,
        )

    async def upsert_relationships_in_bulk_of_same_operation(
        self,
        shape: OperationOnRelationshipIdentity,
        relationships: Iterable[RelationshipWithNodes],
    ):
        batched_query = (
            self.ingest_query_builder.generate_batch_update_relationship_query_batch(
                shape, relationships
            )
        )
        await self.execute(
            batched_query.as_query(self.ingest_query_builder.apoc_iterate),
            log_result=True,
        )

    async def upsert_key_index(self, index: KeyIndex):
        query = self.index_query_builder.create_key_index_query(index)
        await self.execute(query)

    async def upsert_field_index(self, index: FieldIndex):
        query = self.index_query_builder.create_field_index_query(index)
        await self.execute(query)

    async def perform_ttl_op(self, config: TimeToLiveConfiguration):
        query = self.ingest_query_builder.generate_ttl_query_from_configuration(config)
        await self.execute(query)

    async def execute_hook(self, hook: IngestionHook):
        query_string, params = hook.as_cypher_query_and_parameters()
        await self.execute(Query(query_string, params))

    async def execute(self, query: Query, log_result: bool = False):
        self.logger.debug(
            "Executing Cypher Query to Neo4j",
            extra={
                "query": query.query_statement,
                "uri": self.driver._pool.address.host,
            },
        )
        await self.driver.verify_connectivity()
        result = await self.driver.execute_query(
            query.query_statement,
            query.parameters,
            database_=self.database_name,
        )
        if log_result:
            for record in result.records:
                self.logger.info(
                    "Gathered Query Results",
                    extra=dict(**record, query=query.query_statement),
                )



from typing import List

from cleo.helpers import option

from ...project import Project, Target
from ...schema.schema import GraphObjectShape
from ..operations import InitializeProject, RunCopy
from .nodestream_command import NodestreamCommand
from .shared_options import JSON_OPTION, PROJECT_FILE_OPTION


class UnknownTargetError(Exception):
    pass


class Copy(NodestreamCommand):
    name = "copy"
    description = "Copy Data from one target to another"
    options = [
        PROJECT_FILE_OPTION,
        JSON_OPTION,
        option("from", "f", "The target to copy from", flag=False),
        option("to", "t", "The target to copy to", flag=False),
        option("all", "a", "Copy all node and relationship types", flag=True),
        option(
            "node",
            description="Specify a node type to copy",
            flag=False,
            multiple=True,
        ),
        option(
            "relationship",
            description="Specify a relationship type to copy",
            flag=False,
            multiple=True,
        ),
    ]

    async def handle_async(self):
        project = await self.run_operation(InitializeProject())

        try:
            from_target = self.get_taget_from_user(project, "from")
            to_target = self.get_taget_from_user(project, "to")
            schema = project.get_schema()
            all_node_types = schema.known_node_types()
            all_rel_types = schema.known_relationship_types()
            node_types = self.get_type_selection_from_user(all_node_types, "node")
            rel_types = self.get_type_selection_from_user(all_rel_types, "relationship")
        except UnknownTargetError:
            return

        self.line("Starting to Copy:")
        self.line(f"<info>From: {from_target.name}</info>")
        self.line(f"<info>To: {to_target.name}</info>")
        self.line(f"<info>Node Types: {', '.join(node_types)}</info>")
        self.line(f"<info>Relationship Types: {', '.join(rel_types)}</info>")
        await self.run_operation(
            RunCopy(from_target, to_target, project, node_types, rel_types)
        )

    def get_taget_from_user(self, project: Project, action: str) -> Target:
        # If the user has specified the target in the options, we don't need to prompt
        # them for anything. We can just use the target they specified.
        if (choice := self.option(action)) is None:
            prompt = f"Which target would you like to copy {action}?"
            choices = [t for t in project.targets_by_name.keys()]
            choice = self.choice(prompt, choices)

        # If the target they specified is unknown, we should error out.
        if (target := project.get_target_by_name(choice)) is None:
            self.line_error(f"Unknown target: {choice}")
            raise UnknownTargetError

        return target

    def get_type_selection_from_user(
        self, types: List[GraphObjectShape], type_name: str
    ) -> List[str]:
        choices = [str(t.object_type) for t in types]

        # If the user has specified the --all flag, we don't need to prompt them for
        # anything. We can just return all the types.
        if self.option("all"):
            return choices

        # If the user has specified type(s) in the options, we don't need to prompt
        # them for anything. We can just return the type they specified. If they
        # specified an unknown type, we should error out.
        selections_from_options = self.option(type_name)
        if selections_from_options:
            for selection in selections_from_options:
                if selection not in choices:
                    self.line_error(
                        f"Unknown {type_name} type: {selection}. "
                        f"Valid options are: {', '.join(choices)}"
                    )
                    raise UnknownTargetError
            return selections_from_options

        # If the user has not specified the type(s) in the options, we need to prompt
        # them for the type(s). We can just return the type they specified.
        return self.choice(
            f"Which {type_name} types would you like to copy? (You can select multiple by separating them with a comma)",
            choices,
            multiple=True,
        )



from ...project import Project
from ..commands.nodestream_command import NodestreamCommand
from .operation import Operation


class RemovePipelineFromProject(Operation):
    def __init__(
        self,
        project: Project,
        scope_name: str,
        pipeline_name: str,
    ) -> None:
        self.project = project
        self.pipeline_name = pipeline_name
        self.scope_name = scope_name

    async def perform(self, _: NodestreamCommand):
        self.project.delete_pipeline(self.scope_name, self.pipeline_name)



from .lowercase_strings import LowercaseStrings
from .normalizer import InvalidFlagError, Normalizer
from .remove_trailing_dots import RemoveTrailingDots
from .trim_whitespace import TrimWhitespace

__all__ = (
    "Normalizer",
    "LowercaseStrings",
    "RemoveTrailingDots",
    "InvalidFlagError",
    "TrimWhitespace",
)



from abc import ABC, abstractmethod
from functools import cache
from typing import Any

from ...pluggable import Pluggable
from ...subclass_registry import MissingFromRegistryError, SubclassRegistry

NORMALIZER_REGISTRY = SubclassRegistry()


class InvalidFlagError(ValueError):
    """Raised when a normalization flag is not valid."""

    def __init__(self, flag_name, *args: object) -> None:
        super().__init__(
            f"Normalization flag with name '{flag_name}' is not valid.`", *args
        )


@NORMALIZER_REGISTRY.connect_baseclass
class Normalizer(Pluggable, ABC):
    """A `Normalizer` is responsible for turning objects into a consistent form.

    When data is extracted from pipeline records from a value provider, the `Normalizer`
    is responsible for "cleaning up" the raw data such that is consistent. Often this comes
    in with regard to strings.
    """

    entrypoint_name = "normalizers"

    @classmethod
    def setup(cls):
        pass

    @classmethod
    def normalize_by_args(cls, value: Any, **normalizer_args) -> Any:
        for flag_name, enabled in normalizer_args.items():
            if enabled:
                value = cls.by_flag_name(flag_name).normalize_value(value)

        return value

    @classmethod
    def argument_flag(cls):
        return f"do_{NORMALIZER_REGISTRY.name_for(cls)}"

    @classmethod
    @cache
    def by_flag_name(cls, flag_name: str) -> "Normalizer":
        if not flag_name.startswith("do_"):
            raise InvalidFlagError(flag_name)

        try:
            normalizer_class = NORMALIZER_REGISTRY.get(flag_name.strip("do_"))
        except MissingFromRegistryError:
            raise InvalidFlagError(flag_name)

        return normalizer_class()

    @abstractmethod
    def normalize_value(self, value: Any) -> Any:
        raise NotImplementedError



from abc import ABC
from typing import Type

from yaml import SafeLoader

from ...pluggable import Pluggable
from ...subclass_registry import SubclassRegistry

ARGUMENT_RESOLVER_REGISTRY = SubclassRegistry()


@ARGUMENT_RESOLVER_REGISTRY.connect_baseclass
class ArgumentResolver(Pluggable, ABC):
    """An `ArgumentResolver` is a class that can resolve a value by injecting into the yaml parser."""

    entrypoint_name = "argument_resolvers"

    @classmethod
    def install_yaml_tag(cls, loader: Type[SafeLoader]):
        pass



from pathlib import Path
from typing import Dict, Iterable, List, Optional, Tuple, Type, TypeVar

from yaml import SafeLoader

from ..file_io import LoadsFromYamlFile, SavesToYamlFile
from ..pipeline import Step
from ..pipeline.argument_resolvers.argument_resolver import ArgumentResolver
from ..pipeline.scope_config import ScopeConfig
from ..pluggable import Pluggable
from ..schema.schema import (
    AggregatedIntrospectiveIngestionComponent,
    GraphSchema,
    IntrospectiveIngestionComponent,
)
from .pipeline_definition import PipelineDefinition
from .pipeline_scope import PipelineScope
from .run_request import RunRequest
from .target import Target

T = TypeVar("T", bound=Step)


class Project(
    AggregatedIntrospectiveIngestionComponent, LoadsFromYamlFile, SavesToYamlFile
):
    """A `Project` represents a collection of pipelines.

    A project is the top-level object in a nodestream project.
    It contains a collection of scopes, which in turn contain a collection of pipelines.
    When interacting with nodestream programmatically, you will typically interact with a project object.
    This is where pipeline execution begins and where all data about the project is stored.
    """

    @classmethod
    def get_loader(cls) -> Type[SafeLoader]:
        """Get the YAML loader to use when reading this object from a file.

        Returns:
            The YAML loader to use when reading this object from a file.
        """
        NodestreamProjectFileSafeLoader.configure()
        return NodestreamProjectFileSafeLoader

    @classmethod
    def describe_yaml_schema(cls):
        from schema import Optional, Schema

        return Schema(
            {
                Optional("scopes"): {
                    str: PipelineScope.describe_yaml_schema(),
                },
                Optional("plugin_config"): {
                    str: ScopeConfig.describe_yaml_schema(),
                },
                Optional("targets"): {
                    str: {str: object},
                },
            }
        )

    @classmethod
    def from_file_data(cls, data) -> "Project":
        """Creates a project from file data.

        The file data should be a dictionary containing a "scopes" key,
        which should be a dictionary mapping scope names to scope file data and
        should be validated by the schema returned by `describe_yaml_schema`.

        This method should not be called directly. Instead, use `NodestreamProjectFileSafeLoader.load_from_yaml_file`.

        Args:
            data (Dict): The file data to create the project from.

        Returns:
            Project: The project created from the file data.
        """
        scopes_data = data.pop("scopes", {})
        scopes = [
            PipelineScope.from_file_data(*scope_data)
            for scope_data in scopes_data.items()
        ]

        plugins = data.pop("plugin_config", {})
        plugin_cfgs = {
            name: ScopeConfig.from_file_data(value) for name, value in plugins.items()
        }

        targets = data.pop("targets", {})
        target_cfgs = {name: Target(name, value) for name, value in targets.items()}

        project = cls(scopes, plugin_cfgs, target_cfgs)
        for plugin in ProjectPlugin.all():
            plugin().activate(project)
        return project

    def to_file_data(self) -> dict:
        """Returns the file data for the project.

        This method should not be called directly. Instead, use `SavesToYamlFile.save_to_yaml_file`.

        Returns:
            Dict: The file data for the project.
        """
        return {
            "scopes": {
                scope.name: scope.to_file_data()
                for scope in self.scopes_by_name.values()
                if scope.persist
            },
        }

    def __init__(
        self,
        scopes: List[PipelineScope],
        plugin_configs: Dict[str, ScopeConfig] = None,
        targets: Dict[str, Target] = None,
    ):
        self.scopes_by_name: Dict[str, PipelineScope] = {}
        self.plugin_configs = plugin_configs
        self.targets_by_name = targets
        for scope in scopes:
            self.add_scope(scope)

    def get_target_by_name(self, target_name: str) -> Optional[Target]:
        """Returns the target with the given name.

        Args:
            target_name (str): The name of the target to return.

        Returns:
            Optional[Target]: The target with the given name, or None if no target was found.
        """
        return self.targets_by_name.get(target_name)

    async def run(self, request: RunRequest) -> int:
        """Takes a run request and runs the appropriate pipeline.

        Args:
            request (RunRequest): The run request to run.
        """
        pipelines_ran = 0
        for scope in self.scopes_by_name.values():
            pipelines_ran += await scope.run_request(request)
        return pipelines_ran

    async def get_snapshot_for(self, pipeline_name: str) -> list:
        """Returns the output of the pipeline with the given name to be used as a snapshot.

        This method is intended for testing purposes only. It will run the pipeline and return the results.
        The pipeline is run with the `test` annotation, so components that you want to run must be annotated with `test` or not annotated at all.

        Args:
            pipeline_name (str): The name of the pipeline to get a snapshot for.

        Returns:
            str: The snapshot of the pipeline.
        """
        run_request = RunRequest.for_testing(pipeline_name, results := [])
        await self.run(run_request)
        return results

    def add_scope(self, scope: PipelineScope):
        """Adds a scope to the project.

        Args:
            scope (PipelineScope): The scope to add.
        """

        if self.plugin_configs and self.plugin_configs.get(scope.name):
            scope.set_configuration(self.plugin_configs[scope.name])
        self.scopes_by_name[scope.name] = scope

    def get_scopes_by_name(self, scope_name: Optional[str]) -> Iterable[PipelineScope]:
        """Returns the scopes with the given name.

        If `scope_name` is None, all scopes will be returned. If no scopes are found, an empty list will be returned.
        """
        if scope_name is None:
            return self.scopes_by_name.values()

        if scope_name not in self.scopes_by_name:
            return []

        return [self.scopes_by_name[scope_name]]

    def get_all_pipeline_names(self) -> Iterable[str]:
        """Returns all pipeline names in the project."""
        for scope in self.scopes_by_name.values():
            for pipeline in scope.pipelines_by_name.values():
                yield pipeline.name

    def delete_pipeline(
        self,
        scope_name: Optional[str],
        pipeline_name: str,
        remove_pipeline_file: bool = True,
        missing_ok: bool = True,
    ):
        """Deletes a pipeline from the project.

        Args:
            scope_name (Optional[str]): The name of the scope containing the pipeline. If None, all scopes will be searched.
            pipeline_name (str): The name of the pipeline to delete.
            remove_pipeline_file (bool, optional): Whether to remove the pipeline file from disk. Defaults to True.
            missing_ok (bool, optional): Whether to raise an error if the pipeline is not found. Defaults to True.

        Raises:
            ValueError: If the pipeline is not found and `missing_ok` is False.
        """
        for scopes in self.get_scopes_by_name(scope_name):
            scopes.delete_pipeline(
                pipeline_name,
                remove_pipeline_file=remove_pipeline_file,
                missing_ok=missing_ok,
            )

    def get_schema(self, type_overrides_file: Optional[Path] = None) -> GraphSchema:
        """Returns a `GraphSchema` representing the project.

        If a `type_overrides_file` is provided, the schema will be updated with the overrides.
        Since nodestream does not know about the types of the data in the project, this is the
        only way to provide type information to the schema where it is not available natively.

        Every pipeline in the project will be loaded and introspected to build the schema.

        Args:
            type_overrides_file (Optional[Path], optional): A path to a YAML file containing type overrides. Defaults to None.

        Returns:
            GraphSchema: The schema representing the project.
        """
        schema = self.generate_graph_schema()
        if type_overrides_file is not None:
            schema.apply_type_overrides_from_file(type_overrides_file)
        return schema

    def all_subordinate_components(self) -> Iterable[IntrospectiveIngestionComponent]:
        return self.scopes_by_name.values()

    def dig_for_step_of_type(
        self, step_type: Type[T]
    ) -> Iterable[Tuple[PipelineDefinition, int, T]]:
        """Yields all steps of the given type in the project.

        Yields tuples of (pipeline_definition, step_index, step).

        Args:
            step_type (Type[T]): The type of step to look for.

        Yields:
            Tuple[PipelineDefinition, int, T]: The steps of the given type and their locations.
        """
        for scope in self.scopes_by_name.values():
            for pipeline_definition in scope.pipelines_by_name.values():
                pipeline_steps = (
                    pipeline_definition.initialize_for_introspection().steps
                )
                for idx, step in enumerate(pipeline_steps):
                    if isinstance(step, step_type):
                        yield pipeline_definition, idx, step


class NodestreamProjectFileSafeLoader(SafeLoader):
    """A YAML loader that can load nodestream project files.""" ""

    was_configured = False

    @classmethod
    def configure(cls):
        if cls.was_configured:
            return
        for argument_resolver in ArgumentResolver.all():
            argument_resolver.install_yaml_tag(cls)

        cls.was_configured = True


class ProjectPlugin(Pluggable):
    """A plugin that can be used to modify a project.

    Plugins are used to add functionality to projects. They are typically used to add new pipeline scopes
    or to modify existing scopes and pipelines. Plugins are activated when a project is loaded from disk.

    Plugins are registered by subclassing `ProjectPlugin` and implementing the `activate` method.
    Additionally, they must be use the `entry_points` mechanism to register themselves as a plugin.
    """

    entrypoint_name = "projects"

    def activate(self, project: Project):
        """Called when a project is loaded from disk.

        This is where the plugin should modify the project. For example, a plugin might add a new scope
        or modify an existing scope. The plugin should not modify the project file on disk. Instead, it
        should modify the project object in memory.
        """
        raise NotImplementedError



from importlib import import_module
from typing import Optional, Type

DECLARATIVE_INIT_METHOD_NAME = "from_file_data"


class InvalidClassPathError(ValueError):
    """Raised when a class path is invalid."""

    pass


class PipelineComponentInitializationError(ValueError):
    """Raised when a component fails to initialize."""

    def __init__(self, initializer, init_arguments, *args: object) -> None:
        super().__init__(
            "Failed to Initialize Component in Declarative Pipeline. Likely the arguments are incorrect.",
            *args,
        )
        self.initializer = initializer
        self.init_arguments = init_arguments


def find_class(class_path):
    try:
        module_name, class_name = class_path.split(":")
        module = import_module(module_name)
        return getattr(module, class_name)
    except ValueError as e:
        raise InvalidClassPathError(
            f"Class path '{class_path}' is not in the correct format."
        ) from e
    except ImportError as e:
        raise InvalidClassPathError(
            f"Module '{module_name}' could not be imported."
        ) from e
    except AttributeError as e:
        raise InvalidClassPathError(
            f"Class '{class_name}' does not exist in module '{module_name}'."
        ) from e


class ClassLoader:
    """Loads a class from a string path and instantiates it with the given arguments."""

    def __init__(self, class_constratint: Optional[Type] = None) -> None:
        self.class_constratint = class_constratint or object

    def find_class_initializer(self, implementation, factory=None):
        class_definition = find_class(implementation)
        factory_method = factory or DECLARATIVE_INIT_METHOD_NAME
        if hasattr(class_definition, factory_method):
            return getattr(class_definition, factory_method)
        else:
            return class_definition

    def load_class(self, implementation, arguments=None, factory=None):
        arguments = arguments or {}
        initializer = self.find_class_initializer(implementation, factory)
        try:
            result = initializer(**arguments)
        except TypeError as e:
            raise PipelineComponentInitializationError(initializer, arguments) from e

        if not isinstance(result, self.class_constratint):
            raise TypeError(
                f"Expected class of type {self.class_constratint}, but got {type(result)}."
            )
        return result



from logging import getLogger
from typing import Any, Dict, Optional

from neo4j import RoutingControl

from ...pipeline.extractors import Extractor
from .database_connector import Neo4jDatabaseConnector


class Neo4jExtractor(Extractor):
    @classmethod
    def from_file_data(
        cls,
        query: str,
        parameters: Optional[Dict[str, Any]] = None,
        limit: int = 100,
        **connector_args
    ):
        connector = Neo4jDatabaseConnector.from_file_data(**connector_args)
        return cls(query, connector, parameters, limit)

    def __init__(
        self,
        query: str,
        connector: Neo4jDatabaseConnector,
        parameters: Optional[Dict[str, Any]] = None,
        limit: int = 100,
    ) -> None:
        self.connector = connector
        self.query = query
        self.parameters = parameters or {}
        self.limit = limit
        self.logger = getLogger(self.__class__.__name__)

    async def extract_records(self):
        # TODO: In the future, we should extract the database pagination logic from
        # this class and move it to a GraphDatabaseExtractor class following the lead
        # we have of the writer class.
        offset = 0
        should_continue = True
        driver = self.connector.driver
        database_name = self.connector.database_name

        while should_continue:
            params = dict(**self.parameters, limit=self.limit, offset=offset)
            self.logger.info(
                "Running query on neo4j",
                extra=dict(query=self.query, params=params),
            )
            query_results, _, _ = await driver.execute_query(
                self.query,
                params,
                routing_=RoutingControl.READ,
                database_=database_name,
            )
            returned_records = list(query_results)
            should_continue = len(returned_records) > 0
            offset += self.limit
            for item in returned_records:
                yield item



from itertools import chain
from typing import List

from ...model import TimeToLiveConfiguration
from ...pipeline.extractors.ttls import TimeToLiveConfigurationExtractor
from ..project import Project
from .audit import Audit


class AuditTimeToLiveConfigurations(Audit):
    name = "ttls"
    description = "Audit the project for missing TTLs"

    async def get_all_ttl_configurations(
        self, project: Project
    ) -> List[TimeToLiveConfiguration]:
        ttl_extractors = project.dig_for_step_of_type(TimeToLiveConfigurationExtractor)
        return [
            ttl
            for _, _, extractor in ttl_extractors
            async for ttl in extractor.extract_records()
        ]

    async def run(self, project: Project):
        schema = project.generate_graph_schema()
        unused_ttls = await self.get_all_ttl_configurations(project)
        for object_shape in chain(
            schema.known_node_types(), schema.known_relationship_types()
        ):
            for_shape = (ttl for ttl in unused_ttls if ttl.is_for_shape(object_shape))
            ttl = next(for_shape, None)
            if ttl is not None:
                unused_ttls.remove(ttl)
            else:
                err = f"Time to live not configured for {object_shape.graph_object_type} '{object_shape.object_type}'"
                self.failure(err)

        if unused_ttls:
            for ttl in unused_ttls:
                warn = f"Time to live configured for unknown object type '{ttl.object_type}'"
                self.warning(warn)

        if self.failure_count == 0:
            self.success("All Object Types Have TTLs Configured")



from abc import ABC, abstractmethod
from typing import Any, Dict, Iterable, Optional, Tuple

from ...compat import deprecated_arugment
from ...model import Node, NodeCreationRule, PropertySet, Relationship
from ...pipeline.normalizers import LowercaseStrings
from ...pipeline.value_providers import (
    ProviderContext,
    StaticValueOrValueProvider,
    ValueProvider,
)
from ...schema.indexes import FieldIndex, KeyIndex
from ...schema.schema import (
    Cardinality,
    GraphObjectShape,
    GraphObjectType,
    KnownTypeMarker,
    PresentRelationship,
    PropertyMetadataSet,
    UnknownTypeMarker,
)
from ..record_decomposers import RecordDecomposer
from .interpretation import Interpretation

DEFAULT_KEY_NORMALIZATION_ARGUMENTS = {LowercaseStrings.argument_flag(): True}


class InvalidKeyLengthError(ValueError):
    """Raised when a related nodes have differing lengths of key parts returned from a value provider.."""

    def __init__(self, district_lengths, *args: object) -> None:
        lengths = f"({','.join((str(length) for length in district_lengths))})"
        error = f"Node Relationships do not have a consistent key length. Lengths are: ({lengths}) "
        super().__init__(error, *args)


class RelatedNodeKeySearchAlgorithm(ABC):
    def __init__(
        self, key_searches: Dict[str, ValueProvider], key_normalization: Dict[str, Any]
    ) -> None:
        self.node_key = key_searches
        self.key_normalization = key_normalization

    @abstractmethod
    def get_related_node_key_sets(
        self, context: ProviderContext
    ) -> Iterable[Dict[str, Any]]:
        raise NotImplementedError


class SingleNodeKeySearchAlgorithm(RelatedNodeKeySearchAlgorithm):
    def get_related_node_key_sets(
        self, context: ProviderContext
    ) -> Iterable[Dict[str, Any]]:
        return [
            {
                k: v.normalize_single_value(context, **self.key_normalization)
                for k, v in self.node_key.items()
            }
        ]


class MultiNodeKeySearchAlgorithm(RelatedNodeKeySearchAlgorithm):
    def get_related_node_key_sets(self, context: ProviderContext):
        # If we do not have the same length, then there is an error because
        # we do not have pairs to create keys for each node based off of.
        all_values_by_key_property = {
            k: tuple(v.normalize_many_values(context, **self.key_normalization))
            for k, v in self.node_key.items()
        }
        distinct_lengths = {len(vals) for vals in all_values_by_key_property.values()}
        all_of_same_length = len(distinct_lengths) == 1
        if not all_of_same_length:
            raise InvalidKeyLengthError(distinct_lengths)

        # Now that we know everything has the same length, we can simply return the pairs
        # for each index in the arrays.
        common_length = next(iter(distinct_lengths))
        return [
            {k: v[i] for k, v in all_values_by_key_property.items()}
            for i in range(common_length)
        ]


class RelationshipInterpretation(Interpretation, alias="relationship"):
    """Provides a generic method by which to interpret a relationship between the source node and zero-to-many related nodes."""

    __slots__ = (
        "can_find_many",
        "outbound",
        "node_creation_rule",
        "decomposer",
        "node_type",
        "relationship_type",
        "node_key",
        "node_properties",
        "relationship_key",
        "relationship_properties",
        "key_normalization",
        "properties_normalization",
        "key_search_algorithm",
    )

    @deprecated_arugment("match_strategy", "node_creation_rule")
    def __init__(
        self,
        node_type: StaticValueOrValueProvider,
        relationship_type: StaticValueOrValueProvider,
        node_key: Dict[str, StaticValueOrValueProvider],
        node_properties: Optional[Dict[str, StaticValueOrValueProvider]] = None,
        relationship_key: Optional[Dict[str, StaticValueOrValueProvider]] = None,
        relationship_properties: Optional[Dict[str, StaticValueOrValueProvider]] = None,
        outbound: bool = True,
        find_many: bool = False,
        iterate_on: Optional[ValueProvider] = None,
        node_creation_rule: Optional[str] = None,
        key_normalization: Optional[Dict[str, Any]] = None,
        properties_normalization: Optional[Dict[str, Any]] = None,
    ):
        self.can_find_many = find_many or iterate_on is not None
        self.outbound = outbound
        self.node_creation_rule = NodeCreationRule(
            node_creation_rule or NodeCreationRule.EAGER.value
        )
        self.decomposer = RecordDecomposer.from_iteration_arguments(iterate_on)
        self.node_type = ValueProvider.guarantee_value_provider(node_type)
        self.relationship_type = ValueProvider.guarantee_value_provider(
            relationship_type
        )
        self.node_key = ValueProvider.guarantee_provider_dictionary(node_key)
        self.node_properties = ValueProvider.guarantee_provider_dictionary(
            node_properties or {}
        )
        self.relationship_key = ValueProvider.guarantee_provider_dictionary(
            relationship_key or {}
        )
        self.relationship_properties = ValueProvider.guarantee_provider_dictionary(
            relationship_properties or {}
        )
        self.key_normalization = {
            **DEFAULT_KEY_NORMALIZATION_ARGUMENTS,
            **(key_normalization or {}),
        }
        self.properties_normalization = properties_normalization or {}

        key_search_algorithm = (
            MultiNodeKeySearchAlgorithm if find_many else SingleNodeKeySearchAlgorithm
        )
        self.key_search_algorithm = key_search_algorithm(
            self.node_key, self.key_normalization
        )

    def interpret(self, context: ProviderContext):
        for sub_context in self.decomposer.decompose_record(context):
            for relationship, related_node in self.find_matches(sub_context):
                context.desired_ingest.add_relationship(
                    related_node, relationship, self.outbound, self.node_creation_rule
                )

    def find_relationship(self, context: ProviderContext) -> Relationship:
        rel = Relationship(type=self.relationship_type.single_value(context))
        rel.key_values.apply_providers(
            context, self.relationship_key, **self.key_normalization
        )
        rel.properties.apply_providers(
            context, self.relationship_properties, **self.properties_normalization
        )
        return rel

    def find_related_nodes(self, context: ProviderContext) -> Iterable[Node]:
        for key_set in self.key_search_algorithm.get_related_node_key_sets(context):
            node = Node(
                type=self.node_type.single_value(context),
                key_values=PropertySet(key_set),
            )
            if node.has_valid_id:
                node.properties.apply_providers(
                    context, self.node_properties, **self.properties_normalization
                )
                yield node

    def find_matches(
        self, context: ProviderContext
    ) -> Iterable[Tuple[Relationship, Node]]:
        relationship = self.find_relationship(context)
        for related_node in self.find_related_nodes(context):
            yield relationship, related_node

    # NOTE: We cannot participate in introspection when we don't know the relationship
    # or related node type until runtime. Sometimes we can partially participate if we know
    # one or the other.

    def gather_present_relationships(self):
        if not all((self.relationship_type.is_static, self.node_type.is_static)):
            return

        relationship_type = KnownTypeMarker(self.relationship_type.value)
        source_node = UnknownTypeMarker.source_node()
        related_node = KnownTypeMarker(type=self.node_type.value)
        from_type = source_node if self.outbound else related_node
        to_type = related_node if self.outbound else source_node
        foreign_cardinality = Cardinality.MANY
        source_cardinality = (
            Cardinality.MANY if self.can_find_many else Cardinality.SINGLE
        )

        if self.outbound:
            to_side_cardinality, from_side_cardinality = (
                foreign_cardinality,
                source_cardinality,
            )
        else:
            from_side_cardinality, to_side_cardinality = (
                foreign_cardinality,
                source_cardinality,
            )

        yield PresentRelationship(
            from_object_type=from_type,
            to_object_type=to_type,
            relationship_type=relationship_type,
            to_side_cardinality=to_side_cardinality,
            from_side_cardinality=from_side_cardinality,
        )

    def gather_used_indexes(self):
        if self.node_type.is_static:
            related_node_type = self.node_type.value
            yield FieldIndex.for_ttl_timestamp(related_node_type)

            # If we are matching fuzzy or MATCH_ONLY, we cannot rely on the key index
            # to find the related node.
            # TODO: Perhaps in the future we can do a FieldIndex instead?
            if self.node_creation_rule == NodeCreationRule.EAGER:
                yield KeyIndex(related_node_type, frozenset(self.node_key.keys()))

        if self.relationship_type.is_static:
            relationship_type = self.relationship_type.value
            yield FieldIndex.for_ttl_timestamp(
                relationship_type, object_type=GraphObjectType.RELATIONSHIP
            )

    def gather_object_shapes(self):
        if self.node_type.is_static:
            yield GraphObjectShape(
                graph_object_type=GraphObjectType.NODE,
                object_type=KnownTypeMarker(self.node_type.value),
                properties=PropertyMetadataSet.from_names(
                    self.node_properties.keys(), self.node_key.keys()
                ),
            )

        if self.relationship_type.is_static:
            yield GraphObjectShape(
                graph_object_type=GraphObjectType.RELATIONSHIP,
                object_type=KnownTypeMarker(self.relationship_type.value),
                properties=PropertyMetadataSet.from_names(
                    self.relationship_key.keys(), self.relationship_properties.keys()
                ),
            )



import asyncio
from dataclasses import asdict, dataclass, field
from logging import getLogger
from typing import TYPE_CHECKING, List

from .creation_rules import NodeCreationRule, RelationshipCreationRule
from .graph_objects import Node, Relationship, RelationshipWithNodes
from .ingestion_hooks import IngestionHook, IngestionHookRunRequest

if TYPE_CHECKING:
    from ..databases.ingest_strategy import IngestionStrategy


LOGGER = getLogger(__name__)


@dataclass(slots=True)
class DesiredIngestion:
    source: Node = field(default_factory=Node)
    relationships: List[Relationship] = field(default_factory=list)
    hook_requests: List[IngestionHookRunRequest] = field(default_factory=list)

    @property
    def source_node_is_valid(self) -> bool:
        return self.source.is_valid

    async def ingest_source_node(self, strategy: "IngestionStrategy"):
        await strategy.ingest_source_node(self.source)

    async def ingest_relationships(self, strategy: "IngestionStrategy"):
        await asyncio.gather(
            *(
                strategy.ingest_relationship(relationship)
                for relationship in self.relationships
            )
        )

    async def run_ingest_hooks(self, strategy: "IngestionStrategy"):
        await asyncio.gather(
            *(strategy.run_hook(hook_req) for hook_req in self.hook_requests)
        )

    def can_perform_ingest(self):
        # We can do the main part of the ingest if the source node is valid.
        # If its not valid, its only an error when there are relationships we are
        # trying to ingest as well.
        if not self.source_node_is_valid:
            if len(self.relationships) > 0:
                LOGGER.warning(
                    "Identity value for source node was null. Skipping Ingest.",
                    extra=asdict(self),
                )
            else:
                LOGGER.debug(
                    "Ingest was not provided a valid source node and no relationships. Only running ingest hooks.",
                    extra=asdict(self),
                )
            return False
        return True

    async def ingest(self, strategy: "IngestionStrategy"):
        if self.can_perform_ingest():
            await self.ingest_source_node(strategy)
            await self.ingest_relationships(strategy)
        await self.run_ingest_hooks(strategy)

    def add_relationship(
        self,
        related_node: Node,
        relationship: Relationship,
        outbound: bool,
        node_creation_rule: NodeCreationRule = NodeCreationRule.EAGER,
        relationship_creation_rule: RelationshipCreationRule = RelationshipCreationRule.EAGER,
    ):
        if not related_node.is_valid:
            LOGGER.warning(
                "Identity value for related node was null. Skipping.",
                extra=asdict(related_node),
            )
            return

        from_node, to_node = (
            (self.source, related_node) if outbound else (related_node, self.source)
        )
        from_match, to_match = (
            (NodeCreationRule.EAGER, node_creation_rule)
            if outbound
            else (node_creation_rule, NodeCreationRule.EAGER)
        )
        self.relationships.append(
            RelationshipWithNodes(
                from_node=from_node,
                to_node=to_node,
                relationship=relationship,
                from_side_node_creation_rule=from_match,
                to_side_node_creation_rule=to_match,
                relationship_creation_rule=relationship_creation_rule,
            )
        )

    def add_ingest_hook(self, hook: IngestionHook, before_ingest=False):
        self.hook_requests.append(IngestionHookRunRequest(hook, before_ingest))



from typing import Type

from yaml import SafeLoader

from .argument_resolver import ArgumentResolver


class IncludeFileResolver(ArgumentResolver):
    """An `IncludeFileResolver` is an `ArgumentResolver` that can resolve a file path into a file's contents."""

    @classmethod
    def install_yaml_tag(cls, loader: Type[SafeLoader]):
        loader.add_constructor(
            "!include",
            lambda loader, node: cls.include_file(loader.construct_scalar(node)),
        )

    @staticmethod
    def include_file(file_path: str):
        from ..pipeline_file_loader import PipelineFileSafeLoader

        return PipelineFileSafeLoader.load_file_by_path(file_path)



import asyncio
from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor
from logging import getLogger
from typing import Any, AsyncGenerator, Optional

from ..flush import Flush
from ..step import Step


class Transformer(Step):
    """A `Transformer` takes a given record and mutates into a new record.

    `Transformer` steps generally make up the middle of an ETL pipeline and are responsible
    for reshaping an object so its more ingestible by the downstream sink.
    """

    async def handle_async_record_stream(
        self, record_stream: AsyncGenerator[Any, Any]
    ) -> AsyncGenerator[Any, Any]:
        async for record in record_stream:
            if record is Flush:
                yield record
            else:
                val_or_gen = self.transform_record(record)
                if isinstance(val_or_gen, AsyncGenerator):
                    async for result in val_or_gen:
                        yield result
                else:
                    yield await val_or_gen

    @abstractmethod
    async def transform_record(self, record: Any) -> Any:
        raise NotImplementedError


class ConcurrentTransformer(Transformer):
    """A `ConcurrentTransformer` takes a given record and mutates into a new record while managing a pool of concurrent workers.

    `ConcurrentTransformer`s are useful when the transformation work is IO bound and can be parallelized across multiple threads.
    `ConcurrentTransformer`s are not useful when the transformation work is CPU bound as the GIL will prevent useful parallelization.
    Additionally, `ConcurrentTransformer`s do not guarantee the order of the output stream will match the input stream.
    """

    def __init__(
        self, thread_pool_size: Optional[int] = None, maximum_pending_tasks: int = 1000
    ) -> None:
        self.logger = getLogger(self.__class__.__name__)
        self.maximum_pending_tasks = maximum_pending_tasks
        self.thread_pool = ThreadPoolExecutor(
            max_workers=thread_pool_size,
            thread_name_prefix=self.__class__.__name__,
            initializer=self.on_worker_thread_start,
        )

    def on_worker_thread_start(self):
        pass

    async def handle_async_record_stream(
        self, record_stream: AsyncGenerator[Any, Any]
    ) -> AsyncGenerator[Any, Any]:
        pending_tasks = []

        def drain_completed_tasks():
            tasks_drained = 0
            remaining_pending_tasks = []
            for task in pending_tasks:
                if task.done():
                    yield task.result()
                    tasks_drained += 1
                else:
                    remaining_pending_tasks.append(task)

            pending_tasks[:] = remaining_pending_tasks
            if tasks_drained:
                self.logger.debug(
                    "Drained %s completed tasks, %s pending tasks remain",
                    tasks_drained,
                    len(pending_tasks),
                )

        async for record in record_stream:
            if record is Flush:
                # Flush the pending tasks, then yield the flush.
                # In order to fully respect the Flush,
                # we need to wait for all pending tasks to complete.
                while pending_tasks:
                    for result in drain_completed_tasks():
                        yield result
                yield record
            else:
                # Submit the work to the thread pool (only if we have capacity)
                # If we don't have capacity, yield completed tasks until we do.
                submitted = False
                while not submitted:
                    if len(pending_tasks) < self.maximum_pending_tasks:
                        task = self.thread_pool.submit(self.do_work_on_record, record)
                        pending_tasks.append(task)
                        submitted = True
                    for result in drain_completed_tasks():
                        yield result

        # After we've finished enqueuing records, we need to drain all tasks.
        # Items yielded from this loop are the final results of the step.
        self.logger.debug("Finished enqueuing records, draining pending tasks")
        while pending_tasks:
            for result in drain_completed_tasks():
                yield result

    async def finish(self):
        self.thread_pool.shutdown(wait=True)

    def do_work_on_record(self, record: Any) -> Any:
        # Handles the work nessary to transform a single record.
        # This method wraps the `transform_record` method in an asyncio loop.
        # This is necessary because the thread pool executor is not async.
        return asyncio.run(self.transform_record(record))

    @abstractmethod
    async def transform_record(self, record: Any) -> Any:
        raise NotImplementedError



import json
from typing import List, Union

from ...model import JsonLikeDocument
from .transformer import Transformer


class ExpandJsonField(Transformer):
    @classmethod
    def from_file_data(cls, /, path: Union[str, List[str]]):
        if isinstance(path, str):
            path = [path]

        return cls(path)

    def __init__(self, path: List[str]) -> None:
        self.path = path

    async def transform_record(self, record: JsonLikeDocument):
        item = record
        for path_segment in self.path[:-1]:
            record = item[path_segment]
        last_segment = self.path[-1]
        item[last_segment] = json.loads(item[last_segment])
        return record



from abc import ABC, abstractmethod
from typing import Any, AsyncGenerator


class Step(ABC):
    """A `Step` represents a phase of an ETl pipeline."""

    @classmethod
    def from_file_data(cls, **kwargs):
        return cls(**kwargs)

    @abstractmethod
    async def handle_async_record_stream(
        self, record_stream: AsyncGenerator[Any, Any]
    ) -> AsyncGenerator[Any, Any]:
        raise NotImplementedError

    async def finish(self):
        pass


class PassStep(Step):
    """A `PassStep` is a step that does nothing."""

    async def handle_async_record_stream(
        self, record_stream: AsyncGenerator[Any, Any]
    ) -> AsyncGenerator[Any, Any]:
        async for record in record_stream:
            yield record



from typing import Any, Iterable, Type

from yaml import SafeLoader

from .context import ProviderContext
from .value_provider import ValueProvider


class SplitValueProvider(ValueProvider):
    def __init__(self, delimiter: str, data: ValueProvider):
        self.delimiter = delimiter
        self.data = data

    def single_value(self, context: ProviderContext) -> Any:
        return next(self.many_values(context))

    def many_values(self, context: ProviderContext) -> Iterable[Any]:
        for value in self.data.many_values(context):
            if not isinstance(value, str):
                raise TypeError(
                    f"Expected {self.data} to yield strings, but got {value!r}"
                )
            yield from value.split(self.delimiter)

    @classmethod
    def install_yaml_tag(cls, loader: Type[SafeLoader]):
        loader.add_constructor(
            "!split", lambda loader, node: cls(**loader.construct_mapping(node))
        )



from importlib import resources
from typing import Dict, Iterable, List

from ..file_io import LoadsFromYaml, SavesToYaml
from ..pipeline.scope_config import ScopeConfig
from ..schema.schema import (
    AggregatedIntrospectiveIngestionComponent,
    IntrospectiveIngestionComponent,
)
from .pipeline_definition import PipelineDefinition
from .run_request import RunRequest


class MissingExpectedPipelineError(ValueError):
    pass


class PipelineScope(
    AggregatedIntrospectiveIngestionComponent, LoadsFromYaml, SavesToYaml
):
    """A `PipelineScope` represents a collection of pipelines subordinate to a project."""

    def __init__(
        self,
        name: str,
        pipelines: List[PipelineDefinition],
        persist: bool = True,
        config: ScopeConfig = None,
    ) -> None:
        self.persist = persist
        self.name = name
        self.config = config
        self.pipelines_by_name: Dict[str, PipelineDefinition] = {}
        for pipeline in pipelines:
            self.add_pipeline_definition(pipeline)

    @classmethod
    def from_file_data(cls, scope_name, file_data):
        pipelines_data = file_data.pop("pipelines", [])
        annotations = file_data.pop("annotations", {})
        config = file_data.pop("config", {})
        pipelines = [
            PipelineDefinition.from_file_data(pipeline_data, annotations)
            for pipeline_data in pipelines_data
        ]
        return cls(scope_name, pipelines, config=ScopeConfig.from_file_data(config))

    @classmethod
    def describe_yaml_schema(cls):
        from schema import Optional, Or, Schema

        return Schema(
            {
                Optional("pipelines"): [
                    PipelineDefinition.describe_yaml_schema(),
                ],
                Optional("annotations"): {
                    str: Or(str, int, float, bool),
                },
                Optional("config"): ScopeConfig.describe_yaml_schema(),
            }
        )

    def to_file_data(self):
        return {
            "pipelines": [ppl.to_file_data() for ppl in self.pipelines_by_name.values()]
        }

    def all_subordinate_components(self) -> Iterable[IntrospectiveIngestionComponent]:
        return self.pipelines_by_name.values()

    async def run_request(self, run_request: "RunRequest") -> int:
        """Execute a run request.

        If the pipeline does not exist, this is a no-op. Otherwise, the run request
        will be executed with the pipeline's definition. The run request will be
        executed asynchronously via its `execute_with_definition` method.

        Args:
            run_request: The run request to execute.
        """
        if (name := run_request.pipeline_name) not in self:
            return 0

        await run_request.execute_with_definition(self[name], self.config)
        return 1

    def __getitem__(self, pipeline_name):
        return self.pipelines_by_name[pipeline_name]

    def __contains__(self, pipeline_name):
        return pipeline_name in self.pipelines_by_name

    def add_pipeline_definition(self, definition: PipelineDefinition):
        """Add a pipeline to the scope.

        Args:
            definition: The pipeline to add.
        """
        self.pipelines_by_name[definition.name] = definition

    def delete_pipeline(
        self,
        pipeline_name: str,
        remove_pipeline_file: bool = True,
        missing_ok: bool = True,
    ) -> bool:
        """Delete a pipeline from the scope.

        Args:
            pipeline_name: The name of the pipeline to delete.
            remove_pipeline_file: Whether to remove the pipeline's file from disk.
            missing_ok: Whether to raise an error if the pipeline does not exist.

        Returns:
            True if the pipeline was deleted, False if it did not exist.
        """
        definition = self.pipelines_by_name.get(pipeline_name)
        if definition is None:
            if not missing_ok:
                raise MissingExpectedPipelineError(
                    "Attempted to delete pipeline that did not exist"
                )
            else:
                return False

        del self.pipelines_by_name[pipeline_name]
        if remove_pipeline_file:
            definition.remove_file(missing_ok=missing_ok)

        return True

    def set_configuration(self, config):
        self.config = config

    @classmethod
    def from_resources(
        cls, name: str, package: resources.Package, persist: bool = False
    ) -> "PipelineScope":
        """Load a `PipelineScope` from a package's resources.

        Each `.yaml` file in the package's resources will be loaded as a pipeline.
        Internally, this uses `importlib.resources` to load the files and calls
        `PipelineDefinition.from_path` on each file. This means that the name of
        the pipeline will be the name of the file without the .yaml extension.

        Args:
            name: The name of the scope.
            package: The name of the package to load from.
            persist: Whether or not to save the scope when the project is saved.

        Returns:
            A `PipelineScope` instance with the pipelines defined in the package.
        """
        pipelines = [
            PipelineDefinition.from_path(f)
            for f in resources.files(package).iterdir()
            if f.suffix == ".yaml"
        ]
        return cls(name, pipelines, persist=persist)



from ..operations import (
    AddPipelineToProject,
    CommitProjectToDisk,
    GeneratePipelineScaffold,
    InitializeProject,
)
from .nodestream_command import NodestreamCommand
from .shared_options import (
    DATABASE_NAME_OPTION,
    PIPELINE_ARGUMENT,
    PROJECT_FILE_OPTION,
    SCOPE_NAME_OPTION,
)


class Scaffold(NodestreamCommand):
    name = "scaffold"
    description = "Generate a New Nodestream Pipeline"
    arguments = [PIPELINE_ARGUMENT]
    options = [DATABASE_NAME_OPTION, SCOPE_NAME_OPTION, PROJECT_FILE_OPTION]

    async def handle_async(self):
        pipeline_name = self.argument("pipeline")
        desired_scope = self.option("scope")
        database_name = self.option("database")
        project = await self.run_operation(InitializeProject())
        generated_pipelines = await self.run_operation(
            GeneratePipelineScaffold(
                project_root=self.get_project_path().parent,
                database_name=database_name,
                pipeline_file_name=pipeline_name + ".yaml",
            )
        )
        for generated_pipeline in generated_pipelines:
            self.line(f"<info>Generated pipeline at '{generated_pipeline}'</info>")
            await self.run_operation(
                AddPipelineToProject(project, generated_pipeline, desired_scope)
            )

        await self.run_operation(CommitProjectToDisk(project, self.get_project_path()))



from cleo.helpers import argument

from ..operations import (
    CommitProjectToDisk,
    InitializeProject,
    RemovePipelineFromProject,
)
from .nodestream_command import NodestreamCommand
from .shared_options import PIPELINE_ARGUMENT, PROJECT_FILE_OPTION


class Remove(NodestreamCommand):
    name = "remove"
    description = "Remove a Pipeline from Your Nodestream Project"
    arguments = [
        argument("scope", "the name of the scope the pipeline is in"),
        PIPELINE_ARGUMENT,
    ]
    options = [PROJECT_FILE_OPTION]

    async def handle_async(self):
        project = await self.run_operation(InitializeProject())
        remove = RemovePipelineFromProject(
            project, self.argument("scope"), self.argument("pipeline")
        )
        await self.run_operation(remove)
        await self.run_operation(CommitProjectToDisk(project, self.get_project_path()))



from cleo.helpers import option

from ..operations import InitializeProject, PrintProjectSchema
from .nodestream_command import NodestreamCommand
from .shared_options import PROJECT_FILE_OPTION


class PrintSchema(NodestreamCommand):
    name = "print schema"
    description = "Print the schema for the current project"
    options = [
        PROJECT_FILE_OPTION,
        option(
            "format", "f", "Format to print the schema in", default="plain", flag=False
        ),
        option(
            "overrides",
            "O",
            "Path to a file containing type overrides",
            default=None,
            flag=False,
        ),
        option(
            "out",
            "o",
            "Path to a file to write the schema to",
            default=None,
            flag=False,
        ),
    ]

    async def handle_async(self):
        project = await self.run_operation(InitializeProject())
        await self.run_operation(
            PrintProjectSchema(
                project=project,
                format_string=self.option("format"),
                type_overrides_file=self.option("overrides"),
                output_file=self.option("out"),
            )
        )



from importlib.metadata import entry_points
from inspect import getmembers, isclass

NODESTREAM_PLUGINS_ENTRYPOINT_GROUP = "nodestream.plugins"


class Pluggable:
    entrypoint_name: str = "plugin"

    @classmethod
    def entrypoints(cls):
        return entry_points(
            group=NODESTREAM_PLUGINS_ENTRYPOINT_GROUP, name=cls.entrypoint_name
        )

    @classmethod
    def all(cls):
        def is_plugin(member):
            return isclass(member) and issubclass(member, cls) and member is not cls

        for entrypoint in cls.entrypoints():
            plugin_module = entrypoint.load()
            for _, member in getmembers(plugin_module, is_plugin):
                yield member

    @classmethod
    def import_all(cls):
        for _ in cls.all():
            pass



from typing import Any

from ...pipeline.value_providers import ProviderContext, ValueProvider
from .transformer import Transformer


class ValueProjection(Transformer):
    def __init__(self, projection: ValueProvider) -> None:
        self.projection = projection

    async def transform_record(self, record: Any):
        context = ProviderContext.fresh(record)
        for result in self.projection.many_values(context):
            yield result



from functools import wraps
from warnings import warn


def leave_unchanged(v):
    return v


def deprecated_arugment(old, new, converter=leave_unchanged):
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            if old in kwargs:
                kwargs[new] = converter(kwargs[old])
                del kwargs[old]
                warn(
                    f"Argument {old} is deprecated. Use {new} instead.",
                    DeprecationWarning,
                    stacklevel=2,
                )

            return f(*args, **kwargs)

        return wrapper

    return decorator



from typing import AsyncGenerator

from neo4j.graph import Node as Neo4jNode
from neo4j.graph import Relationship as Neo4jRelationship

from ...model import Node, PropertySet, Relationship, RelationshipWithNodes
from ..copy import TypeRetriever
from .database_connector import Neo4jDatabaseConnector
from .extractor import Neo4jExtractor

FETCH_ALL_NODES_BY_TYPE_QUERY_FORMAT = """
MATCH (n:{type})
RETURN n SKIP $offset LIMIT $limit
"""

FETCH_ALL_RELATIONSHIPS_BY_TYPE_QUERY_FORMAT = """
MATCH (a)-[r:{type}]->(b)
RETURN a, r, b SKIP $offset LIMIT $limit
"""


class Neo4jTypeRetriever(TypeRetriever):
    def __init__(self, connector: Neo4jDatabaseConnector) -> None:
        self.connector = connector

    def map_neo4j_node_to_nodestream_node(
        self, node: Neo4jNode, type: str = None
    ) -> Node:
        # NOTE: I don't think this will work in all cases.
        # But I think this will require shaking out in the future.
        type = type or next(iter(node.labels))
        return Node(
            type=type,
            properties=PropertySet(node),
            additional_types=tuple(label for label in node.labels if label != type),
        )

    def map_neo4j_relationship_to_nodestream_relationship(
        self, relationship: Neo4jRelationship
    ) -> Relationship:
        return Relationship(
            type=relationship.type,
            properties=PropertySet(relationship),
        )

    def get_node_type_extractor(self, type: str) -> Neo4jExtractor:
        return Neo4jExtractor(
            FETCH_ALL_NODES_BY_TYPE_QUERY_FORMAT.format(type=type), self.connector
        )

    def get_relationship_type_extractor(self, type: str) -> Neo4jExtractor:
        return Neo4jExtractor(
            FETCH_ALL_RELATIONSHIPS_BY_TYPE_QUERY_FORMAT.format(type=type),
            self.connector,
        )

    async def get_nodes_of_type(self, type: str) -> AsyncGenerator[Node, None]:
        extractor = self.get_node_type_extractor(type)
        async for row in extractor.extract_records():
            yield self.map_neo4j_node_to_nodestream_node(row["n"], type=type)

    async def get_relationships_of_type(
        self, type: str
    ) -> AsyncGenerator[RelationshipWithNodes, None]:
        extractor = self.get_relationship_type_extractor(type)
        async for row in extractor.extract_records():
            yield RelationshipWithNodes(
                from_node=self.map_neo4j_node_to_nodestream_node(row["a"]),
                to_node=self.map_neo4j_node_to_nodestream_node(row["b"]),
                relationship=self.map_neo4j_relationship_to_nodestream_relationship(
                    row["r"]
                ),
            )



from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Dict, List, Optional

from yaml import SafeLoader, load

from .argument_resolvers import ArgumentResolver
from .class_loader import ClassLoader
from .normalizers import Normalizer
from .pipeline import Pipeline
from .scope_config import ScopeConfig
from .step import Step
from .value_providers import ValueProvider


class InvalidPipelineDefinitionError(ValueError):
    """Raised when a pipeline definition is invalid."""

    pass


class PipelineFileSafeLoader(SafeLoader):
    """A YAML loader that can load pipeline files.""" ""

    was_configured = False

    @classmethod
    def configure(cls, config: ScopeConfig = None):
        if config:
            cls.add_constructor(
                "!config",
                lambda loader, node: config.get_config_value(
                    loader.construct_scalar(node)
                ),
            )

        if cls.was_configured:
            return

        for normalizer in Normalizer.all():
            normalizer.setup()
        for value_provider in ValueProvider.all():
            value_provider.install_yaml_tag(cls)
        for argument_resolver in ArgumentResolver.all():
            argument_resolver.install_yaml_tag(cls)

        cls.was_configured = True

    @classmethod
    def load_file_by_path(cls, file_path: str, config: ScopeConfig = None):
        PipelineFileSafeLoader.configure(config)
        with open(file_path) as fp:
            return load(fp, cls)


@dataclass(slots=True)
class PipelineInitializationArguments:
    """Arguments used to initialize a pipeline from a file."""

    step_outbox_size: int = 1000
    annotations: Optional[List[str]] = None
    on_effective_configuration_resolved: Optional[Callable[[List[Dict]], None]] = None
    extra_steps: Optional[List[Step]] = None

    @classmethod
    def for_introspection(cls):
        return cls(annotations=["introspection"])

    @classmethod
    def for_testing(cls):
        return cls(annotations=["test"])

    def initialize_from_file_data(self, file_data: List[dict]):
        return Pipeline(
            steps=self.load_steps(ClassLoader(Step), file_data),
            step_outbox_size=self.step_outbox_size,
        )

    def load_steps(self, class_loader, file_data):
        effective = self.get_effective_configuration(file_data)
        if self.on_effective_configuration_resolved:
            self.on_effective_configuration_resolved(file_data)
        in_file = [class_loader.load_class(**step_data) for step_data in effective]
        return in_file + (self.extra_steps or [])

    def get_effective_configuration(self, file_data):
        return [
            step_data for step_data in file_data if self.should_load_step(step_data)
        ]

    def should_load_step(self, step):
        return self.step_is_tagged_properly(step)

    def step_is_tagged_properly(self, step):
        # By default, we need to load all steps. Only filter a step out if:
        #   1. We have set annotations during initialization (e.g from the cli)
        #   2. The step is annotated
        #   3. There is not an intersection between the annotations from (1) and (2).
        #
        # This function also has a side effect of removing the annotations from the step data.
        annotations_set_on_step = set(step.pop("annotations", []))
        if annotations_set_on_step and self.annotations:
            return annotations_set_on_step.intersection(self.annotations)

        return True


class PipelineFileLoader:
    """Loads a pipeline from a YAML file."""

    def __init__(self, file_path: Path):
        self.file_path = file_path

    def load_pipeline(
        self,
        init_args: Optional[PipelineInitializationArguments] = None,
        config: ScopeConfig = None,
    ) -> Pipeline:
        init_args = init_args or PipelineInitializationArguments()
        return self.load_pipeline_from_file_data(
            self.load_pipeline_file_data(config), init_args
        )

    def load_pipeline_from_file_data(
        self, file_data, init_args: PipelineInitializationArguments
    ):
        if not isinstance(file_data, list):
            raise InvalidPipelineDefinitionError(
                "File should be a list of step class to load"
            )

        return init_args.initialize_from_file_data(file_data)

    def load_pipeline_file_data(self, config: ScopeConfig = None):
        return PipelineFileSafeLoader.load_file_by_path(self.file_path, config)



import re

from jinja2 import Environment

from ..schema import (
    Cardinality,
    GraphObjectShape,
    GraphSchema,
    PresentRelationship,
    PropertyType,
)
from .schema_printer import SchemaPrinter

NODE_TYPE_DEF_TEMPLATE = """
{% set node_type = ensure_camel_case(shape.object_type.type) %}
type {{ node_type }} @exclude(operations: [CREATE, DELETE, UPDATE]) @pageOptions{limit: {default: 10}} {% if node_type != shape.object_type.type %} @node(labels: ["{{ shape.object_type.type }}"]) {% endif %} {
    # Node Properties
{% for property in shape.properties.properties.values() %}
    {{ property.name }}: {{ field_mappings[property.type] }}
{% endfor %}
    # Inbound Relationships
{% for rel in inbound_relationships %}
    {% if rel.to_side_cardinality == Cardinality.MANY %}
    {{ in_relationship_to_camel(rel.relationship_type, rel.from_object_type) }}: [{{ ensure_camel_case(rel.from_object_type.type) }}!]!
        @relationship(type: "{{ rel.relationship_type }}", direction: IN, properties: "{{ unique_rel_type_name(rel) }}")
    {% else %}
    {{ in_relationship_to_camel(rel.relationship_type, rel.from_object_type) }}: {{ ensure_camel_case(rel.from_object_type.type) }}
        @relationship(type: "{{ rel.relationship_type }}", direction: IN, properties: "{{ unique_rel_type_name(rel) }}")
    {% endif %}
{% endfor %}
    # Outbound Relationships
{% for rel in outbound_relationships %}
    {% if rel.from_side_cardinality == Cardinality.MANY %}
    {{ out_relationship_to_camel(rel.relationship_type, rel.to_object_type) }}: [{{ensure_camel_case( rel.to_object_type.type) }}!]!
        @relationship(type: "{{ rel.relationship_type }}", direction: OUT, properties: "{{ unique_rel_type_name(rel) }}")
    {% else %}
    {{ out_relationship_to_camel(rel.relationship_type, rel.to_object_type) }}: {{ensure_camel_case( rel.to_object_type.type) }}
        @relationship(type: "{{ rel.relationship_type }}", direction: OUT, properties: "{{ unique_rel_type_name(rel) }}")
    {% endif %}
{% endfor %}
}
"""

RELATIONSHIP_DEF_TEMPLATE = """
interface {{ unique_rel_type_name_from_shape(rel) }} @relationshipProperties {
{% for property in rel.properties.properties.values() %}
    {{ property.name }}: {{ field_mappings[property.type] }}
{% endfor %}
}
"""


SNAKE_CASE_REGEX = re.compile(r"^[a-z0-9]+(_[a-z0-9]+)*$")

PROPERTY_TYPE_TO_GRAPQHL_TYPES = {
    PropertyType.STRING: "String",
    PropertyType.DATETIME: "LocalDateTime",
    PropertyType.INTEGER: "BigInt",
    PropertyType.BOOLEAN: "Boolean",
    PropertyType.FLOAT: "Float",
}


def unique_rel_type_name_from_shape(rel: GraphObjectShape):
    rel_type = rel.object_type.type
    return "".join(p.title() for p in rel_type.lower().split("_"))


def unique_rel_type_name(rel: PresentRelationship):
    rel_type = ensure_camel_case(rel.relationship_type.type)
    return "".join(p.title() for p in rel_type.lower().split("_"))


def ensure_camel_case(possible_snake_case):
    if SNAKE_CASE_REGEX.match(possible_snake_case):
        splits = possible_snake_case.split("_")
        return "".join(t.title() for t in splits)
    return possible_snake_case


def in_relationship_to_camel(relationship, other_node):
    splits = str(relationship).lower().split("_")
    other_node = ensure_camel_case(str(other_node))
    return other_node[0].lower() + other_node[1:] + "".join((t.title() for t in splits))


def out_relationship_to_camel(relationship, other_node):
    splits = str(relationship).lower().split("_")
    other_node = ensure_camel_case(str(other_node))
    return splits[0] + "".join((t.title() for t in splits[1:])) + other_node


class GraphQLSchemaPrinter(SchemaPrinter, alias="graphql"):
    def __init__(self) -> None:
        environment = Environment()
        environment.globals.update(
            {
                item.__name__: item
                for item in (
                    Cardinality,
                    in_relationship_to_camel,
                    out_relationship_to_camel,
                    ensure_camel_case,
                    unique_rel_type_name,
                    unique_rel_type_name_from_shape,
                )
            }
        )

        self.node_template = environment.from_string(NODE_TYPE_DEF_TEMPLATE)
        self.rel_template = environment.from_string(RELATIONSHIP_DEF_TEMPLATE)

    def render_node_schema(self, schema: GraphSchema) -> str:
        representation = ""
        for node_shape in sorted(
            schema.known_node_types(), key=lambda k: k.object_type.type
        ):
            all_inbound_rels = [
                r
                for r in schema.relationships
                if r.to_object_type == node_shape.object_type
            ]
            all_outbound_rels = [
                r
                for r in schema.relationships
                if r.from_object_type == node_shape.object_type
            ]
            representation += self.node_template.render(
                shape=node_shape,
                field_mappings=PROPERTY_TYPE_TO_GRAPQHL_TYPES,
                inbound_relationships=all_inbound_rels,
                outbound_relationships=all_outbound_rels,
            )
        return representation

    def render_relationship_schema(self, schema: GraphSchema) -> str:
        return "".join(
            self.rel_template.render(
                rel=rel, field_mappings=PROPERTY_TYPE_TO_GRAPQHL_TYPES
            )
            for rel in schema.known_relationship_types()
        )

    def print_schema_to_string(self, schema: GraphSchema) -> str:
        representation = self.render_node_schema(schema)
        representation += self.render_relationship_schema(schema)
        return representation + "\n"



import os
from typing import Type

from yaml import SafeLoader

from .argument_resolver import ArgumentResolver


class EnvironmentResolver(ArgumentResolver):
    """An `EnvironmentResolver` is an `ArgumentResolver` that can resolve an environment variable into its value."""

    @classmethod
    def install_yaml_tag(cls, loader: Type[SafeLoader]):
        loader.add_constructor(
            "!env",
            lambda loader, node: cls.resolve_argument(loader.construct_scalar(node)),
        )

    @staticmethod
    def resolve_argument(variable_name):
        return os.environ.get(variable_name)



from abc import ABC, abstractclassmethod, abstractmethod
from pathlib import Path
from typing import Type

from schema import Schema
from yaml import SafeDumper, SafeLoader, dump, load


class DescribesYamlSchema(ABC):
    """A mixin for classes that can be described by a YAML schema."""

    @abstractclassmethod
    def describe_yaml_schema(self) -> Schema:
        raise NotImplementedError


class LoadsFromYaml(DescribesYamlSchema):
    """A mixin for classes that can be read from YAML."""

    @abstractclassmethod
    def from_file_data(cls, data):
        """Create an instance of this class from the given file data.

        Args:
            data: The data read from the file.

        Returns:
            An instance of this class.
        """
        raise NotImplementedError

    @classmethod
    def validate_and_load(cls, data):
        """Validate the given data against the YAML schema and load it into an instance of this class.

        Args:
            data: The data to validate and load.

        Raises:
            SchemaError: If the data does not match the YAML schema.

        Returns:
            An instance of this class.
        """
        validated_data = cls.describe_yaml_schema().validate(data)
        return cls.from_file_data(validated_data)


class LoadsFromYamlFile(LoadsFromYaml):
    """A mixin for classes that can be read from a YAML file."""

    @classmethod
    def get_loader(cls) -> Type[SafeLoader]:
        """Get the YAML loader to use when reading this object from a file.

        Returns:
            The YAML loader to use when reading this object from a file.
        """
        return SafeLoader

    @classmethod
    def read_from_file(cls, file_path: Path) -> "LoadsFromYaml":
        """Read this object from a YAML file.

        Args:
            file_path: The path to the file to read from.

        Raises:
            FileNotFoundError: If the file does not exist.

        Returns:
            An instance of this class.
        """
        if not file_path.exists():
            raise FileNotFoundError(f"File '{file_path}' does not exist.")

        with file_path.open("r") as f:
            file_data = load(f, Loader=cls.get_loader())
            return cls.validate_and_load(file_data)


class SavesToYaml(DescribesYamlSchema):
    """A mixin for classes that can be written to as YAML."""

    @abstractmethod
    def to_file_data(self):
        """Get the data to write to a YAML file.

        Returns:
            The data to write to a YAML file.
        """
        raise NotImplementedError


class SavesToYamlFile(SavesToYaml):
    """A mixin for classes that can be written to a YAML file."""

    @classmethod
    def get_dumper(cls) -> Type[SafeDumper]:
        """Get the YAML dumper to use when writing this object to a file.

        Returns:
            The YAML dumper to use when writing this object to a file.
        """
        return SafeDumper

    def write_to_file(self, file_path: Path):
        """Write this object to a YAML file.

        If the file already exists, it will be overwritten.

        Args:
            file_path: The path to the file to write to.

        Returns:
            None
        """
        file_data = self.to_file_data()
        validated_file_data = self.describe_yaml_schema().validate(file_data)
        with file_path.open("w") as f:
            dump(
                validated_file_data,
                f,
                Dumper=self.get_dumper(),
                indent=2,
                sort_keys=True,
            )



from abc import abstractmethod
from typing import Any, AsyncGenerator

from ..step import Step


class Extractor(Step):
    """Extractors represent the source of a set of records.

    They are like any other step. However, they ignore the incoming record stream and instead produce their own
    stream of records. For this reason they generally should only be set at the beginning of a pipeline.
    """

    def handle_async_record_stream(
        self, _: AsyncGenerator[Any, Any]
    ) -> AsyncGenerator[Any, Any]:
        return self.extract_records()

    @abstractmethod
    async def extract_records(self) -> AsyncGenerator[Any, Any]:
        raise NotImplementedError



from .argument_resolver import ARGUMENT_RESOLVER_REGISTRY, ArgumentResolver
from .environment_variable_resolver import EnvironmentResolver
from .include_file_resolver import IncludeFileResolver

__all__ = (
    "ARGUMENT_RESOLVER_REGISTRY",
    "ArgumentResolver",
    "EnvironmentResolver",
    "IncludeFileResolver",
)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Iterable

from ..model import (
    IngestionHook,
    Node,
    NodeCreationRule,
    NodeIdentityShape,
    RelationshipCreationRule,
    RelationshipIdentityShape,
    RelationshipWithNodes,
    TimeToLiveConfiguration,
)
from ..schema.indexes import FieldIndex, KeyIndex


@dataclass(slots=True, frozen=True)
class OperationOnNodeIdentity:
    node_identity: NodeIdentityShape
    node_creation_rule: NodeCreationRule


@dataclass(slots=True, frozen=True)
class OperationOnRelationshipIdentity:
    from_node: OperationOnNodeIdentity
    to_node: OperationOnNodeIdentity
    relationship_identity: RelationshipIdentityShape
    relationship_creation_rule: RelationshipCreationRule


class QueryExecutor(ABC):
    @abstractmethod
    async def upsert_nodes_in_bulk_with_same_operation(
        self, operation: OperationOnNodeIdentity, nodes: Iterable[Node]
    ):
        raise NotImplementedError

    @abstractmethod
    async def upsert_relationships_in_bulk_of_same_operation(
        self,
        shape: OperationOnRelationshipIdentity,
        relationships: Iterable[RelationshipWithNodes],
    ):
        raise NotImplementedError

    @abstractmethod
    async def upsert_key_index(self, index: KeyIndex):
        raise NotImplementedError

    @abstractmethod
    async def upsert_field_index(self, index: FieldIndex):
        raise NotImplementedError

    @abstractmethod
    async def perform_ttl_op(self, config: TimeToLiveConfiguration):
        raise NotImplementedError

    @abstractmethod
    async def execute_hook(self, hook: IngestionHook):
        raise NotImplementedError



from abc import abstractmethod
from logging import INFO, getLevelName, getLogger
from typing import Any, AsyncGenerator

from .step import Step


class Writer(Step):
    """A `Writer` takes a given record and commits it to a downstream data store.

    `Writer` steps generally make up the end of an ETL pipeline and are responsible
    for ensuring that the newly transformed data is persisted. After writing the record,
    the record is passed downstream.
    """

    async def handle_async_record_stream(
        self, record_stream: AsyncGenerator[Any, Any]
    ) -> AsyncGenerator[Any, Any]:
        async for record in record_stream:
            await self.write_record(record)
            yield record

    @abstractmethod
    async def write_record(self, record: Any):
        raise NotImplementedError


class LoggerWriter(Writer):
    """A `Writer` that logs the record to a logger."""

    def __init__(self, logger_name=None, level=INFO) -> None:
        logger_name = logger_name or self.__class__.__name__
        self.logger = getLogger(logger_name)
        # At first glance this line would appear wrong. `getLevelName` does the
        # mapping bi-directionally. If you give it a number (level), you get the name
        # and vice-versa. Since we are standardizing on the int representation of
        # the level (thats what Logger#log requires), then we need to
        # call `getLevelName` when the value is the string name for the log level.
        #
        # see: https://docs.python.org/3/library/logging.html#logging.getLevelName
        self.level = getLevelName(level) if isinstance(level, str) else level

    async def write_record(self, record):
        self.logger.log(self.level, record)



from pathlib import Path
from typing import Any, Iterable

import yaml

from ...pipeline.value_providers import JmespathValueProvider
from ..commands.nodestream_command import NodestreamCommand
from .operation import Operation

DEFAULT_PIPELINE_FILE_NAME = "sample.yaml"

SIMPLE_PIPELINE = [
    {
        "implementation": "nodestream.pipeline:IterableExtractor",
        "factory": "range",
        "arguments": {"stop": 100000},
    },
    {
        "implementation": "nodestream.interpreting:Interpreter",
        "arguments": {
            "interpretations": [
                {
                    "type": "source_node",
                    "node_type": "Number",
                    "key": {
                        "number": JmespathValueProvider.from_string_expression("index")
                    },
                }
            ]
        },
    },
]


def pretty_print_yaml_to_file(path: Path, data: Any):
    with open(path, "w") as fp:
        yaml.safe_dump(data, fp, indent=2, sort_keys=True)


class GeneratePipelineScaffold(Operation):
    def __init__(
        self,
        project_root: Path,
        database_name: str,
        pipeline_file_name: str = DEFAULT_PIPELINE_FILE_NAME,
    ) -> None:
        self.project_root = project_root
        self.database_name = database_name
        self.pipeline_file_name = pipeline_file_name

    async def perform(self, _: NodestreamCommand) -> Iterable[Path]:
        path = self.prepare_file_path()
        self.make_pipeline_at_path(path)
        return [path]

    def prepare_file_path(self) -> Path:
        pipeline_dir = self.project_root / "pipelines"
        pipeline_dir.mkdir(parents=True, exist_ok=True)
        return pipeline_dir / self.pipeline_file_name

    def make_pipeline_at_path(self, path: Path):
        pretty_print_yaml_to_file(path, SIMPLE_PIPELINE)



from enum import Enum


class NodeCreationRule(str, Enum):
    EAGER = "EAGER"
    MATCH_ONLY = "MATCH_ONLY"
    FUZZY = "FUZZY"

    def prevent_creation(self) -> "NodeCreationRule":
        if self == NodeCreationRule.EAGER:
            return NodeCreationRule.MATCH_ONLY
        return self


class RelationshipCreationRule(str, Enum):
    EAGER = "EAGER"
    CREATE = "CREATE"



from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

from ..schema.schema import GraphObjectShape, GraphObjectType, KnownTypeMarker

if TYPE_CHECKING:
    from ..databases.ingest_strategy import IngestionStrategy


@dataclass(frozen=True, slots=True)
class TimeToLiveConfiguration:
    graph_object_type: GraphObjectType
    object_type: str
    expiry_in_hours: Optional[int] = 24
    custom_query: Optional[str] = None
    batch_size: int = 100
    enabled: bool = True

    async def ingest(self, strategy: "IngestionStrategy"):
        if self.enabled:
            await strategy.perform_ttl_operation(self)

    def is_for_shape(self, shape: GraphObjectShape):
        return (
            self.graph_object_type == shape.graph_object_type
            and KnownTypeMarker(self.object_type) == shape.object_type
        )



from .expand_json_field import ExpandJsonField
from .transformer import ConcurrentTransformer, Transformer
from .value_projection import ValueProjection

__all__ = ("ExpandJsonField", "ValueProjection", "Transformer", "ConcurrentTransformer")



from functools import wraps


class AlreadyInRegistryError(ValueError):
    """Raised when a subclass with the same name is already in the subclass registry."""

    def __init__(self, name, *args: object) -> None:
        super().__init__(f"{name} is already registered", *args)


class MissingFromRegistryError(ValueError):
    """Raised when a subclass is not in the subclass registry."""

    def __init__(self, name, *args: object) -> None:
        super().__init__(f"{name} is not in the subclass registry", *args)


class SubclassRegistry:
    """A registry for subclasses of a base class."""

    __slots__ = ("registry", "linked_base")

    def __init__(self) -> None:
        self.registry = {}
        self.linked_base = None

    def connect_baseclass(self, base_class):
        """Connect a base class to this registry."""

        old_init_subclass = base_class.__init_subclass__
        base_class.__subclass__registry__ = self
        self.linked_base = base_class

        @classmethod
        @wraps(old_init_subclass)
        def init_subclass(cls, alias=None, *args, **kwargs):
            alias = alias or cls.__name__

            if alias in self.registry:
                raise AlreadyInRegistryError(alias)

            self.registry[alias] = cls
            return old_init_subclass(*args, **kwargs)

        base_class.__init_subclass__ = init_subclass
        return base_class

    def name_for(self, cls):
        """Get the name of a class in the registry."""

        for k, v in self.registry.items():
            if v is cls:
                return k

        return None

    def get(self, name):
        """Get a subclass by name."""
        try:
            return self.registry[name]
        except KeyError:
            raise MissingFromRegistryError(name)

    @property
    def all_subclasses(self):
        return self.registry.values()



from pathlib import Path

from ...pluggable import Pluggable
from ...subclass_registry import SubclassRegistry
from ..schema import GraphSchema

SCHEMA_PRINTER_SUBCLASS_REGISTRY = SubclassRegistry()


@SCHEMA_PRINTER_SUBCLASS_REGISTRY.connect_baseclass
class SchemaPrinter(Pluggable):
    entrypoint_name = "schema_printers"

    def print_schema_to_file(self, schema: GraphSchema, file_path: Path):
        with open(file_path, "w") as f:
            f.write(self.print_schema_to_string(schema))

    def print_schema_to_stdout(self, schema: GraphSchema, print_fn=print):
        print_fn(self.print_schema_to_string(schema))

    def print_schema_to_string(self, schema: GraphSchema) -> str:
        return str(schema)

    @classmethod
    def from_name(cls, name: str) -> "SchemaPrinter":
        return SCHEMA_PRINTER_SUBCLASS_REGISTRY.get(name)()



import json
from abc import ABC, abstractmethod
from typing import Iterable, List, Optional, Tuple

from ...project import PipelineDefinition, Project
from ..commands.nodestream_command import NodestreamCommand
from .operation import Operation


class OutputFormat(ABC):
    def __init__(self, command: NodestreamCommand) -> None:
        self.command = command

    @abstractmethod
    def output(self, matching_pipelines: Iterable[Tuple[str, PipelineDefinition]]):
        raise NotImplementedError


class TableOutputFormat(OutputFormat):
    HEADERS = ["scope", "name", "file", "annotations"]

    def get_data_rows(
        self, matching_pipelines: Iterable[Tuple[str, PipelineDefinition]]
    ) -> List[list]:
        return [
            [
                scope,
                definition.name,
                str(definition.file_path),
                "".join(definition.annotations.keys()),
            ]
            for scope, definition in matching_pipelines
        ]

    def output(self, matching_pipelines: Iterable[Tuple[str, PipelineDefinition]]):
        table = self.command.table(self.HEADERS, self.get_data_rows(matching_pipelines))
        table.render()


class JsonOutputFormat(OutputFormat):
    def output(self, matching_pipelines: Iterable[Tuple[str, PipelineDefinition]]):
        json_data = [
            pipeline.to_file_data(verbose=self.command.is_verbose)
            for _, pipeline in matching_pipelines
        ]
        self.command.write(json.dumps(json_data))


class ShowPipelines(Operation):
    def __init__(
        self, project: Project, scope_name: Optional[str], use_json: bool = False
    ) -> None:
        self.project = project
        self.scope_name = scope_name
        self.use_json = use_json

    async def perform(self, command: NodestreamCommand):
        self.get_output_format(command).output(self.get_matching_pipelines())

    def get_output_format(self, command: NodestreamCommand):
        cls = JsonOutputFormat if self.use_json else TableOutputFormat
        return cls(command)

    def get_matching_pipelines(self) -> Iterable[Tuple[str, PipelineDefinition]]:
        if self.scope_name:
            scopes = [self.project.scopes_by_name[self.scope_name]]
        else:
            scopes = self.project.scopes_by_name.values()

        for scope in scopes:
            for pipeline in scope.pipelines_by_name.values():
                yield scope.name, pipeline



class AuditPrinter:
    """Prints audit messages to an output source. This is the default implementation that uses print to output to the console."""

    def print_success(self, message: str) -> None:
        print(message)

    def print_failure(self, message: str) -> None:
        print(message)

    def print_warning(self, message: str) -> None:
        print(message)



from cookiecutter.main import cookiecutter

from ..commands.nodestream_command import NodestreamCommand
from .operation import Operation

PROJECT_COOKIECUTTER_URL = "gh:nodestream-proj/nodestream-project-cookiecutter"


class RunProjectCookiecutter(Operation):
    def __init__(self, project_name, project_path, database):
        self.project_name = project_name
        self.project_path = project_path
        self.database = database

    async def perform(self, _: NodestreamCommand):
        from ..application import get_version

        cookiecutter(
            PROJECT_COOKIECUTTER_URL,
            no_input=True,
            output_dir=self.project_path,
            extra_context={
                "project_name": self.project_name,
                "database": self.database,
                "nodestream_version": get_version(),
            },
        )



from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Iterable

from ..pipeline import Flush, Step
from ..pipeline.value_providers import ProviderContext
from ..schema.schema import (
    AggregatedIntrospectiveIngestionComponent,
    IntrospectiveIngestionComponent,
)
from .interpretations import Interpretation
from .record_decomposers import RecordDecomposer


class InterpretationPass(IntrospectiveIngestionComponent, ABC):
    @classmethod
    def from_file_data(self, args):
        if args is None:
            return NullInterpretationPass()

        if len(args) > 0 and isinstance(args[0], list):
            return MultiSequenceInterpretationPass.from_file_data(args)

        return SingleSequenceInterpretationPass.from_file_data(args)

    @abstractmethod
    def apply_interpretations(self, context: ProviderContext):
        pass


class NullInterpretationPass(
    AggregatedIntrospectiveIngestionComponent, InterpretationPass
):
    def apply_interpretations(self, context: ProviderContext):
        yield context

    def all_subordinate_components(
        self,
    ) -> Iterable[IntrospectiveIngestionComponent]:
        return []


class MultiSequenceInterpretationPass(
    AggregatedIntrospectiveIngestionComponent, InterpretationPass
):
    __slots__ = ("passes",)

    @classmethod
    def from_file_data(cls, args):
        return cls(*(InterpretationPass.from_file_data(arg) for arg in args))

    def __init__(self, *passes: InterpretationPass) -> None:
        self.passes = passes

    def apply_interpretations(self, context: ProviderContext):
        for interpretation_pass in self.passes:
            provided_subcontext = deepcopy(context)
            for res in interpretation_pass.apply_interpretations(provided_subcontext):
                yield res

    def all_subordinate_components(self) -> Iterable[IntrospectiveIngestionComponent]:
        yield from self.passes


class SingleSequenceInterpretationPass(
    AggregatedIntrospectiveIngestionComponent, InterpretationPass
):
    __slots__ = ("interpretations",)

    @classmethod
    def from_file_data(cls, interpretation_arg_list):
        interpretations = (
            Interpretation.from_file_data(**args) for args in interpretation_arg_list
        )
        return cls(*interpretations)

    def __init__(self, *interpretations: Interpretation):
        self.interpretations = interpretations

    def apply_interpretations(self, context: ProviderContext):
        for interpretation in self.interpretations:
            interpretation.interpret(context)
        yield context

    def all_subordinate_components(self) -> Iterable[IntrospectiveIngestionComponent]:
        yield from self.interpretations


class Interpreter(Step, AggregatedIntrospectiveIngestionComponent):
    __slots__ = (
        "before_iteration",
        "interpretations",
        "decomposer",
    )

    @classmethod
    def from_file_data(cls, interpretations, before_iteration=None, iterate_on=None):
        # Import all interpretation plugins before we try to load any interpretations.
        Interpretation.import_all()

        return cls(
            before_iteration=InterpretationPass.from_file_data(before_iteration),
            interpretations=InterpretationPass.from_file_data(interpretations),
            decomposer=RecordDecomposer.from_iteration_arguments(iterate_on),
        )

    def __init__(
        self,
        before_iteration: InterpretationPass,
        interpretations: InterpretationPass,
        decomposer: RecordDecomposer,
    ) -> None:
        self.before_iteration = before_iteration
        self.interpretations = interpretations
        self.decomposer = decomposer

    async def handle_async_record_stream(self, record_stream):
        # Step 1: Emit any indexes that need to be created.
        # Step 2: Iterate through the stream and emit the appropriate ingestible objects.
        # NOTE: If any record is a flush, do nothing and pass it down stream.
        for index in self.gather_used_indexes():
            yield index

        async for record in record_stream:
            if record is Flush:
                yield record
                continue

            for output_context in self.interpret_record(record):
                yield output_context.desired_ingest

    def interpret_record(self, record):
        context = ProviderContext.fresh(record)
        for base_context in self.before_iteration.apply_interpretations(context):
            for sub_context in self.decomposer.decompose_record(base_context):
                yield from self.interpretations.apply_interpretations(sub_context)

    def all_subordinate_components(self) -> Iterable[IntrospectiveIngestionComponent]:
        yield self.before_iteration
        yield self.interpretations



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, Tuple


class IngestionHook(ABC):
    """An IngestionHook is a custom piece of logic that is bundled as a query.

    IngestionHooks provide a mechanism to enrich the graph model that is derived from
    data added to and currently existing in the graph. For example, drawing an edge
    between two nodes where following a complex path.
    """

    @abstractmethod
    def as_cypher_query_and_parameters(self) -> Tuple[str, Dict[str, Any]]:
        """Returns a cypher query string and parameters to execute."""

        raise NotImplementedError


@dataclass(slots=True, frozen=True)
class IngestionHookRunRequest:
    """An `IngestionHookRunRequest` defines what hook is meant to be run and when the hook should run."""

    hook: IngestionHook
    before_ingest: bool



from typing import Any

from .normalizer import Normalizer


class LowercaseStrings(Normalizer, alias="lowercase_strings"):
    def normalize_value(self, value: Any) -> Any:
        return value.lower() if isinstance(value, str) else value



from .context import ProviderContext
from .jmespath_value_provider import JmespathValueProvider
from .mapping_value_provider import MappingValueProvider
from .regex_value_provider import RegexValueProvider
from .split_value_provider import SplitValueProvider
from .static_value_provider import StaticValueProvider
from .string_format_value_provider import StringFormattingValueProvider
from .value_provider import (
    VALUE_PROVIDER_REGISTRY,
    StaticValueOrValueProvider,
    ValueProvider,
)
from .variable_value_provider import VariableValueProvider

__all__ = (
    "ProviderContext",
    "JmespathValueProvider",
    "MappingValueProvider",
    "SplitValueProvider",
    "StaticValueProvider",
    "StringFormattingValueProvider",
    "ValueProvider",
    "VariableValueProvider",
    "VALUE_PROVIDER_REGISTRY",
    "StaticValueOrValueProvider",
    "RegexValueProvider",
)



from .schema_printer import SchemaPrinter


class PlainTestSchemaPrinter(SchemaPrinter, alias="plain"):
    def print_schema_to_string(self, schema):
        schema_str = "\n".join(str(s) for s in schema.object_shapes if s.has_known_type)
        return f"{schema_str}\n"
